package ru.dbotthepony.mc.otm.menu

import net.minecraft.network.FriendlyByteBuf
import net.minecraft.world.SimpleContainer
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.ItemStack
import net.minecraftforge.energy.CapabilityEnergy
import net.minecraftforge.network.NetworkEvent
import ru.dbotthepony.mc.otm.Registry
import ru.dbotthepony.mc.otm.block.entity.BlockEntityDriveViewer
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.drive.IMatteryDrive
import ru.dbotthepony.mc.otm.item.ItemPortableCondensationDrive
import ru.dbotthepony.mc.otm.item.ItemPortableCondensationDrive.FilterSettings
import ru.dbotthepony.mc.otm.menu.data.INetworkedItemViewSupplier
import ru.dbotthepony.mc.otm.menu.data.NetworkedItemView
import ru.dbotthepony.mc.otm.storage.ItemStackWrapper
import ru.dbotthepony.mc.otm.storage.PoweredVirtualComponent
import java.util.function.Supplier

class DriveViewerMenu @JvmOverloads constructor(
	containerID: Int,
	inventory: Inventory,
	tile: BlockEntityDriveViewer? = null
) : PoweredMatteryMenu(
	Registry.Menus.DRIVE_VIEWER, containerID, inventory, tile
), INetworkedItemViewSupplier {
	@JvmField val view: NetworkedItemView
	private val powered: PoweredVirtualComponent<ItemStackWrapper>?
	@JvmField val driveSlot: MatterySlot
	private var lastDrive: IMatteryDrive<ItemStackWrapper>? = null

	override fun getNetworkedItemView() = view

	init {
		val container = if (tile != null) tile.drive_slot else SimpleContainer(1)

		driveSlot = object : MatterySlot(container, 0) {
			override fun mayPlace(stack: ItemStack): Boolean {
				return stack.getCapability(MatteryCapability.DRIVE).isPresent
			}
		}

		view = NetworkedItemView(inventory.player, this, tile == null)

		if (tile != null) {
			powered = PoweredVirtualComponent(
				ItemStackWrapper::class.java,
				tile.getCapability(MatteryCapability.ENERGY).resolve().get()
			)

			view.setComponent(powered)
		} else {
			powered = null
		}

		addSlot(driveSlot)
		addBatterySlot()
		addInventorySlots()
	}

	override fun broadcastChanges() {
		super.broadcastChanges()

		if (tile != null) {
			val itemStack = (tile as BlockEntityDriveViewer).drive_slot.getItem(0)

			val prev = lastDrive
			lastDrive = null

			if (!itemStack.isEmpty) {
				itemStack.getCapability(MatteryCapability.DRIVE).ifPresent {
					if (it.storageIdentity() == ItemStackWrapper::class.java) {
						lastDrive = it as IMatteryDrive<ItemStackWrapper>
					}
				}
			}

			if (prev != lastDrive) {
				if (prev != null)
					powered!!.remove(prev)

				view.clear()

				if (lastDrive != null)
					powered!!.add(lastDrive!!)
			}

			view.network()
		}
	}

	override fun getWorkingSlotStart(): Int {
		return 0
	}

	override fun getWorkingSlotEnd(): Int {
		return 2
	}

	override fun removed(p_38940_: Player) {
		super.removed(p_38940_)

		if (powered != null)
			lastDrive?.removeListener(powered)

		view.clear()
	}

	override fun quickMoveStack(ply: Player, slot_index: Int): ItemStack {
		val slot = slots[slot_index]
		val item = slot.item

		if (item.isEmpty || item.getCapability(MatteryCapability.DRIVE).isPresent || item.getCapability(CapabilityEnergy.ENERGY).isPresent)
			return super.quickMoveStack(ply, slot_index)

		val powered = powered
		if (lastDrive == null || powered == null)
			return ItemStack.EMPTY

		val remaining = powered.insertStack(ItemStackWrapper(item), false)

		if (remaining.count.toInt() == item.count)
			return ItemStack.EMPTY

		if (remaining.isEmpty()) {
			val copy = item.copy()
			slot.set(ItemStack.EMPTY)
			return copy
		}

		val copy = item.copy()
		item.shrink(remaining.count.toInt())
		slot.setChanged()
		return copy
	}

	enum class FilterSwitch {
		MATCH_NBT, MATCH_TAG, BLACKLIST
	}

	fun getFilter(): FilterSettings? {
		if (driveSlot.item.isEmpty)
			return null

		val item = driveSlot.item.item
		return if (item is ItemPortableCondensationDrive) item.getFilterSettings(driveSlot.item) else null
	}

	class FilterSetPacket(val id: Int, val slot: Int, val value: ItemStack) {
		fun write(buffer: FriendlyByteBuf) {
			buffer.writeInt(id)
			buffer.writeInt(slot)
			buffer.writeItem(value)
		}

		fun play(context: Supplier<NetworkEvent.Context>) {
			context.get().packetHandled = true

			if (slot < 0 || slot >= FilterSettings.MAX_FILTERS)
				return

			if (value.count > 1)
				value.count = 1

			context.get().enqueueWork {
				val ply = context.get().sender!!
				val menu = ply.containerMenu

				if (menu is DriveViewerMenu && menu.containerId == id) {
					if (menu.driveSlot.item.isEmpty)
						return@enqueueWork

					val itemStack = menu.driveSlot.item
					val item = itemStack.item

					if (item is ItemPortableCondensationDrive) {
						val filter = item.getFilterSettings(itemStack)
						filter.items[slot] = value
						filter.serializeNBT(itemStack)
					}
				}
			}
		}

		companion object {
			@JvmStatic
			fun read(buffer: FriendlyByteBuf): FilterSetPacket {
				return FilterSetPacket(buffer.readInt(), buffer.readInt(), buffer.readItem())
			}
		}
	}

	class FilterSwitchPacket(val id: Int, val type: FilterSwitch, val value: Boolean) {
		fun write(buffer: FriendlyByteBuf) {
			buffer.writeInt(id)
			buffer.writeEnum(type)
			buffer.writeBoolean(value)
		}

		fun play(context: Supplier<NetworkEvent.Context>) {
			context.get().packetHandled = true
			context.get().enqueueWork {
				val ply = context.get().sender!!
				val menu = ply.containerMenu

				if (menu is DriveViewerMenu && menu.containerId == id) {
					val settings = menu.getFilter() ?: return@enqueueWork

					when (type) {
						FilterSwitch.MATCH_NBT -> {
							settings.matchNbt = value
						}

						FilterSwitch.MATCH_TAG -> {
							settings.matchTag = value
						}

						FilterSwitch.BLACKLIST -> {
							settings.isBlacklist = value
						}
					}

					settings.serializeNBT(menu.driveSlot.item)
				}
			}
		}

		companion object {
			@JvmStatic
			fun read(buffer: FriendlyByteBuf): FilterSwitchPacket {
				return FilterSwitchPacket(
					buffer.readInt(),
					buffer.readEnum(FilterSwitch::class.java),
					buffer.readBoolean()
				)
			}
		}
	}
}