package ru.dbotthepony.mc.otm.menu.data

import com.mojang.blaze3d.platform.InputConstants
import net.minecraft.client.gui.screens.Screen
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.ClickAction
import net.minecraft.world.inventory.ClickType
import net.minecraft.world.item.ItemStack
import net.minecraftforge.network.PacketDistributor
import ru.dbotthepony.mc.otm.core.Fraction
import ru.dbotthepony.mc.otm.menu.MatteryMenu
import ru.dbotthepony.mc.otm.network.MatteryNetworking
import ru.dbotthepony.mc.otm.network.SetCarriedPacket
import ru.dbotthepony.mc.otm.storage.IStorageComponent
import ru.dbotthepony.mc.otm.storage.IStorageListener
import ru.dbotthepony.mc.otm.storage.IStorageView
import ru.dbotthepony.mc.otm.storage.ItemStackWrapper
import java.util.*

/**
 * Creates a virtual, slotless container for Player's interaction with.
 */
open class NetworkedItemView(val ply: Player, val menu: MatteryMenu, val remote: Boolean) : IStorageListener<ItemStackWrapper> {
	@JvmRecord
	data class NetworkedItem @JvmOverloads constructor(val id: Int, val stack: ItemStack, val upstreamId: UUID? = null)

	@JvmField protected var next_stack_id = 0
	@JvmField val state = HashMap<Int, NetworkedItem>()

	@JvmField protected val upstream_state = HashMap<UUID, NetworkedItem>()
	@JvmField protected val backlog = ArrayList<Any>()

	operator fun get(id: Int): NetworkedItem? {
		return state[id]
	}

	private var cachedView: List<NetworkedItem>? = null
	@JvmField protected var provider: IStorageComponent<ItemStackWrapper>? = null

	fun mouseClick(index: Int, mouse_click_type: Int) {
		val list = getItems()

		val action =
			if (mouse_click_type == InputConstants.MOUSE_BUTTON_LEFT) ClickAction.PRIMARY else ClickAction.SECONDARY

		val type =
			if (mouse_click_type == InputConstants.MOUSE_BUTTON_MIDDLE) ClickType.CLONE else if (Screen.hasShiftDown()) ClickType.QUICK_MOVE else ClickType.PICKUP

		MatteryNetworking.send(
			null,
			InteractPacket(menu.containerId, if (index >= list.size) -1 else list[index].id, type, action)
		)
	}

	fun setComponent(provider: IStorageComponent<ItemStackWrapper>?) {
		if (provider === this.provider) return

		this.provider?.removeListenerAuto(this)
		this.provider = provider
		provider?.addListenerAuto(this)
	}

	fun removed() {
		provider?.removeListenerAuto(this)
	}

	fun clearCache() {
		cachedView = null
	}

	fun getItems(): List<NetworkedItem> {
		return if (cachedView != null) cachedView!! else java.util.List.copyOf(state.values).also { cachedView = it }
	}

	fun getItemCount(): Int {
		return state.values.size
	}

	override fun addObject(stack: ItemStackWrapper, id: UUID, provider: IStorageView<ItemStackWrapper>) = addObject(stack.stack, id)
	override fun changeObject(id: UUID, newCount: Fraction) = changeObject(id, newCount.toInt())

	protected fun network(fn: () -> Any) {
		if (!remote) {
			backlog.add(fn())
		}
	}

	override fun removeObject(id: UUID) {
		val get = upstream_state[id] ?: throw IllegalStateException("Unknown ItemStack with upstream id $id!")
		upstream_state.remove(id)
		state.remove(get.id)
		network { StackRemovePacket(menu.containerId, get.id) }
		clearCache()
	}

	fun addObject(stack: ItemStack, id_upstream: UUID) {
		check(!upstream_state.containsKey(id_upstream)) { "Already tracking ItemStack with upstream id $id_upstream!" }

		val state = NetworkedItem(next_stack_id++, stack.copy(), id_upstream)

		this.state[state.id] = state
		upstream_state[id_upstream] = state
		network { StackAddPacket(menu.containerId, state.id, stack) }
		clearCache()
	}

	fun changeObject(id_upstream: UUID, new_count: Int) {
		val get = upstream_state[id_upstream] ?: throw IllegalStateException("Unknown ItemStack with upstream id $id_upstream!")
		get.stack.count = new_count
		network { StackChangePacket(menu.containerId, get.id, new_count) }
		clearCache()
	}

	fun clear() {
		clearCache()
		upstream_state.clear()
		state.clear()

		if (!remote) {
			backlog.clear()
			backlog.add(ClearPacket(menu.containerId))
		}
	}

	fun network() {
		check(!remote) { "Not a server" }
		val consumer = PacketDistributor.PLAYER.with { ply as ServerPlayer }
		for (packet in backlog) MatteryNetworking.CHANNEL.send(consumer, packet)
		backlog.clear()
	}

	fun playerInteract(packet: InteractPacket) {
		val provider = provider ?: return

		val click = packet.click()
		val action = packet.action()
		val stack_id = packet.stack_id()

		if (click == ClickType.CLONE) {
			if (stack_id < 0 || !ply.abilities.instabuild) return

			val state = get(stack_id) ?: return
			val copy: ItemStack = state.stack.copy()

			copy.count = Math.min(copy.count, copy.maxStackSize)
			ply.containerMenu.carried = copy
			MatteryNetworking.send(ply as ServerPlayer, SetCarriedPacket(ply.containerMenu.carried))
			ply.containerMenu.setRemoteCarried(ply.containerMenu.carried.copy())

			return
		}

		if (click == ClickType.QUICK_MOVE && stack_id > -1) {
			val state = get(stack_id) ?: return

			val amount =
				if (action == ClickAction.PRIMARY)
					state.stack.maxStackSize
				else
					Math.max(1, state.stack.maxStackSize / 2)

			val extracted = provider.extractStack(state.upstreamId!!, amount, true)

			if (!extracted.isEmpty()) {
				val (_, remaining) = menu.quickMoveToInventory(extracted.stack, false)

				if (remaining.count != extracted.stack.count) {
					provider.extractStack(state.upstreamId, extracted.stack.count - remaining.count, false)
				}
			}

			return
		}

		if (!menu.carried.isEmpty && click != ClickType.QUICK_MOVE) {
			// try to put
			if (action == ClickAction.PRIMARY) {
				val carried = menu.carried
				val amount = carried.count

				if (amount == carried.count) {
					val (stack) = provider.insertStack(ItemStackWrapper(menu.carried), false)
					menu.carried = stack
					MatteryNetworking.send(ply as ServerPlayer, SetCarriedPacket(menu.carried))
					menu.setRemoteCarried(menu.carried.copy())
				}
			} else {
				val copy = menu.carried.copy()
				copy.count = 1

				if (provider.insertStack(ItemStackWrapper(copy), false).isEmpty()) {
					menu.carried.shrink(1)
					MatteryNetworking.send(ply as ServerPlayer, SetCarriedPacket(menu.carried))
					menu.setRemoteCarried(menu.carried.copy())
				}
			}
		} else if (stack_id > -1) {
			val state = get(stack_id) ?: return

			val amount =
				if (action == ClickAction.PRIMARY)
					state.stack.maxStackSize
				else
					Math.max(1, state.stack.maxStackSize / 2)

			val extracted = provider.extractStack(state.upstreamId!!, amount, false)
			menu.carried = extracted.stack
			MatteryNetworking.send(ply as ServerPlayer, SetCarriedPacket(menu.carried))
			menu.setRemoteCarried(menu.carried.copy())
		}
	}
}
